# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.waste import Waste


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_push_sewer(self):
        sewer = Sewer(name="", capacity=10, pipe_time=1)

        d1 = {"volume": 4, "phosphate": 2, "temperature": 10}

        d2 = {"volume": 10, "phosphate": 0, "temperature": 0}
        reply = sewer.push_check_sewer()
        self.assertDictAlmostEqual(d2, reply)

        d2["volume"] = 4
        reply = sewer.push_check_sewer(d1)
        self.assertDictAlmostEqual(d2, reply)

        d2["volume"] = 0
        d2["temperature"] = 10  # Related to TODO in nodes.py/QueueTank/push_storage

        reply = sewer.push_set_sewer(d1)
        self.assertDictAlmostEqual(d2, reply)
        self.assertDictAlmostEqual(d1, sewer.sewer_tank.storage)
        self.assertDictAlmostEqual(d1, sewer.sewer_tank.internal_arc.queue[1])

        d2["temperature"] = 0
        self.assertDictAlmostEqual(d2, sewer.sewer_tank.active_storage)

        sewer.end_timestep()
        self.assertDictAlmostEqual(d1, sewer.sewer_tank.active_storage)

    def test_push_land(self):
        sewer = Sewer(name="", capacity=10, pipe_timearea={0: 0.3, 1: 0.7})
        d1 = {"volume": 4, "phosphate": 2, "temperature": 10}
        sewer.push_set_land(d1)

        self.assertDictAlmostEqual(d1, sewer.sewer_tank.storage)

        d2 = {"volume": 4 * 0.3, "phosphate": 2 * 0.3, "temperature": 10}

        d3 = {"volume": 4 * 0.7, "phosphate": 2 * 0.7, "temperature": 10}
        self.assertDictAlmostEqual(d2, sewer.sewer_tank.active_storage)
        self.assertDictAlmostEqual(d3, sewer.sewer_tank.internal_arc.queue[1])

        sewer.end_timestep()
        self.assertDictAlmostEqual(d1, sewer.sewer_tank.active_storage)

    def test_make_discharge(self):
        sewer = Sewer(name="", capacity=10, pipe_timearea={0: 0.3, 1: 0.7})
        waste = Waste(name="")
        arc1 = Arc(in_port=sewer, out_port=waste, capacity=2)
        d1 = {"volume": 9, "phosphate": 2, "temperature": 10}
        sewer.push_set_land(d1)

        sewer.make_discharge()

        d2 = {"volume": 2, "phosphate": 2 * 2 / 9, "temperature": 10}
        self.assertDictAlmostEqual(d2, arc1.vqip_in, 15)

        d3 = {"volume": 7, "phosphate": 7 * 2 / 9, "temperature": 10}
        self.assertDictAlmostEqual(d3, sewer.sewer_tank.storage, 15)

    def test_sewer_overrides(self):
        sewer = Sewer(name="", capacity=10, pipe_timearea={0: 0.3, 1: 0.7})
        sewer.apply_overrides(
            {
                "capacity": 3,
                "chamber_area": 2,
                "chamber_floor": 3.5,
                "pipe_time": 8.4,
                "pipe_timearea": {0: 0.5, 1: 0.5},
            }
        )
        self.assertEqual(sewer.capacity, 3)
        self.assertEqual(sewer.sewer_tank.capacity, 3)
        self.assertEqual(sewer.chamber_area, 2)
        self.assertEqual(sewer.sewer_tank.area, 2)
        self.assertEqual(sewer.chamber_floor, 3.5)
        self.assertEqual(sewer.sewer_tank.datum, 3.5)
        self.assertEqual(sewer.pipe_time, 8.4)
        self.assertEqual(sewer.pipe_timearea, {0: 0.5, 1: 0.5})


if __name__ == "__main__":
    unittest.main()



from wsimod.extensions import register_node_patch


@register_node_patch("my_reservoir", "pull_set_handler", item="FWTW")
def custom_pulls_fwtw(self, vqip, *args, **kwargs):
    """A custom `pull_set_handler` function.

    Pull from the storage when pulled with the tag "FWTW".
    """
    return self.tank.pull_storage(vqip)


@register_node_patch("my_reservoir", "pull_check_handler", item="FWTW")
def custom_pullc_fwtw(self, vqip, *args, **kwargs):
    """A custom `pull_check_handler` function.

    Return available storage when pulled with the tag "FWTW".
    """
    return self.tank.get_avail()


@register_node_patch("my_reservoir", "pull_set_handler", item="default")
def custom_pulls_default(self, vqip, *args, **kwargs):
    """A custom `pull_set_handler` function.

    Deny pull sets by default.
    """
    return self.pull_set_deny(vqip)


@register_node_patch("my_reservoir", "pull_check_handler", item="default")
def custom_pullc_default(self, vqip, *args, **kwargs):
    """A custom `pull_check_handler` function.

    Deny pull checks by default.
    """
    return self.pull_check_deny()



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson

Converted to totals BD 2022-05-03
"""
from typing import Any, Dict

from wsimod.core import constants
from wsimod.nodes.nodes import Node


class Demand(Node):
    """"""

    def __init__(
        self,
        name,
        constant_demand=0,
        pollutant_load={},
        data_input_dict={},
    ):
        """Node that generates and moves water. Currently only subclass
        ResidentialDemand is in use.

        Args:
            name (str): node name constant_demand (float, optional): A constant portion
            of demand if no subclass
                is used. Defaults to 0.
            pollutant_load (dict, optional): Pollutant mass per timestep of
            constant_demand.
                Defaults to 0.
            data_input_dict (dict, optional):  Dictionary of data inputs relevant for
                the node (temperature). Keys are tuples where first value is the name of
                the variable to read from the dict and the second value is the time.
                Defaults to {}

        Functions intended to call in orchestration:
            create_demand
        """
        # TODO should temperature be defined in pollutant dict? TODO a lot of this
        # should be moved to ResidentialDemand Assign parameters
        self.constant_demand = constant_demand
        self.pollutant_load = pollutant_load
        # Update args
        super().__init__(name, data_input_dict=data_input_dict)
        # Update handlers
        self.push_set_handler["default"] = self.push_set_deny
        self.push_check_handler["default"] = self.push_check_deny
        self.pull_set_handler["default"] = self.pull_set_deny
        self.pull_check_handler["default"] = self.pull_check_deny

        # Initialise states
        self.total_demand = self.empty_vqip()
        self.total_backup = self.empty_vqip()  # ew
        self.total_received = self.empty_vqip()

        # Mass balance Because we assume demand is always satisfied received water
        # 'disappears' for mass balance and consumed water 'appears' (this makes)
        # introduction of pollutants easy
        self.mass_balance_in.append(lambda: self.total_demand)
        self.mass_balance_out.append(lambda: self.total_backup)
        self.mass_balance_out.append(lambda: self.total_received)

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the sewer.

        Enables a user to override any of the following parameters:
        constant_demand, pollutant_load.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.constant_demand = overrides.pop("constant_demand", self.constant_demand)
        self.pollutant_load.update(overrides.pop("pollutant_load", {}))
        super().apply_overrides(overrides)

    def create_demand(self):
        """Function to call get_demand, which should return a dict with keys that match
        the keys in directions.

        A dict that determines how to push_distributed the generated wastewater/garden
        irrigation. Water is drawn from attached nodes.
        """
        demand = self.get_demand()
        total_requested = 0
        for dem in demand.values():
            total_requested += dem["volume"]

        self.total_received = self.pull_distributed({"volume": total_requested})

        # TODO Currently just assume all water is received and then pushed onwards
        if (total_requested - self.total_received["volume"]) > constants.FLOAT_ACCURACY:
            print(
                "demand deficit of {2} at {0} on {1}".format(
                    self.name, self.t, total_requested - self.total_received["volume"]
                )
            )

        directions = {
            "garden": {"tag": ("Demand", "Garden"), "of_type": "Land"},
            "house": {"tag": "Demand", "of_type": "Sewer"},
            "default": {"tag": "default", "of_type": None},
        }

        # Send water where it needs to go
        for key, item in demand.items():
            # Distribute
            remaining = self.push_distributed(
                item, of_type=directions[key]["of_type"], tag=directions[key]["tag"]
            )
            self.total_backup = self.sum_vqip(self.total_backup, remaining)
            if remaining["volume"] > constants.FLOAT_ACCURACY:
                print("Demand not able to push")

        # Update for mass balance
        for dem in demand.values():
            self.total_demand = self.sum_vqip(self.total_demand, dem)

    def get_demand(self):
        """Holder function to enable constant demand generation.

        Returns:
            (dict): A VQIP that will contain constant demand
        """
        # TODO read/gen demand
        pol = self.v_change_vqip(self.empty_vqip(), self.constant_demand)
        for key, item in self.pollutant_load.items():
            pol[key] = item
        return {"default": pol}

    def end_timestep(self):
        """Reset state variable trackers."""
        self.total_demand = self.empty_vqip()
        self.total_backup = self.empty_vqip()
        self.total_received = self.empty_vqip()


class NonResidentialDemand(Demand):
    """Holder class to enable non-residential demand generation."""

    def get_demand(self):
        """Holder function.

        Returns:
            (dict): A dict of VQIPs, where the keys match with directions
                in Demand/create_demand
        """
        return {"house": self.get_demand()}


class ResidentialDemand(Demand):
    """"""

    def __init__(
        self,
        population=1,
        pollutant_load={},
        per_capita=0.12,
        gardening_efficiency=0.6 * 0.7,  # Watering efficiency by irrigated area
        data_input_dict={},  # For temperature
        constant_temp=30,
        constant_weighting=0.2,
        **kwargs,
    ):
        """Subclass of demand with functions to handle internal and external water use.

        Args:
            population (float, optional): population of node. Defaults to 1. per_capita
            (float, optional): Volume per person per timestep of water
                used. Defaults to 0.12.
            pollutant_load (dict, optional): Mass per person per timestep of
                different pollutants generated. Defaults to {}.
            gardening_efficiency (float, optional): Value between 0 and 1 that
                translates irrigation demand from GardenSurface into water requested
                from the distribution network. Should account for percent of garden that
                is irrigated and the efficacy of people in meeting their garden water
                demand. Defaults to 0.6*0.7.
            data_input_dict (dict, optional):  Dictionary of data inputs relevant for
                the node (temperature). Keys are tuples where first value is the name of
                the variable to read from the dict and the second value is the time.
                Defaults to {}
            constant_temp (float, optional): A constant temperature associated with
                generated water. Defaults to 30
            constant_weighting (float, optional): Proportion of temperature that is
                made up from by constant_temp. Defaults to 0.2.

        Key assumptions:
            - Per capita calculations to generate demand based on population.
            - Pollutant concentration of generated demand uses a fixed mass per person
              per timestep.
            - Temperature of generated wastewater is based partially on air temperature
              and partially on a constant.
            - Can interact with `land.py/GardenSurface` to simulate garden water use.

        Input data and parameter requirements:
            - `population`.
                _Units_: n
            - `per_capita`.
                _Units_: m3/timestep
            - `data_input_dict` should contain air temperature at model timestep.
                _Units_: C
        """
        self.gardening_efficiency = gardening_efficiency
        self.population = population
        self.per_capita = per_capita
        self.constant_weighting = constant_weighting
        self.constant_temp = constant_temp
        super().__init__(
            data_input_dict=data_input_dict, pollutant_load=pollutant_load, **kwargs
        )
        # Label as Demand class so that other nodes treat it the same
        self.__class__.__name__ = "Demand"

    def apply_overrides(self, overrides: Dict[str, Any] = {}):
        """Apply overrides to the sewer.

        Enables a user to override any of the following parameters:
        gardening_efficiency, population, per_capita, constant_weighting, constant_temp.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        self.gardening_efficiency = overrides.pop(
            "gardening_efficiency", self.gardening_efficiency
        )
        self.population = overrides.pop("population", self.population)
        self.per_capita = overrides.pop("per_capita", self.per_capita)
        self.constant_weighting = overrides.pop(
            "constant_weighting", self.constant_weighting
        )
        self.constant_temp = overrides.pop("constant_temp", self.constant_temp)
        super().apply_overrides(overrides)

    def get_demand(self):
        """Overwrite get_demand and replace with custom functions.

        Returns:
            (dict): A dict of VQIPs, where the keys match with directions
                in Demand/create_demand
        """
        water_output = {}

        water_output["garden"] = self.get_garden_demand()
        water_output["house"] = self.get_house_demand()

        return water_output

    def get_garden_demand(self):
        """Calculate garden water demand in the current timestep by get_connected to all
        attached land nodes. This check should return garden water demand. Applies
        irrigation coefficient. Can function when a single population node is connected
        to multiple land nodes, however, the capacity and preferences of arcs should be
        updated to reflect what is possible based on area.

        Returns:
            vqip (dict): A VQIP of garden water use (including pollutants) to be
                pushed to land
        """
        # Get garden water demand
        excess = self.get_connected(
            direction="push", of_type="Land", tag=("Demand", "Garden")
        )["avail"]

        # Apply garden_efficiency
        excess = self.excess_to_garden_demand(excess)

        # Apply any pollutants
        vqip = self.apply_gardening_pollutants(excess)
        return vqip

    def apply_gardening_pollutants(self, excess):
        """Holder function to apply pollutants (i.e., presumably fertiliser) to the
        garden.

        Args:
            excess (float): A volume of water applied to a garden

        Returns:
            (dict): A VQIP of water that includes pollutants to be sent to land
        """
        # TODO Fertilisers are currently applied in the land node... which is
        # preferable?
        vqip = self.empty_vqip()
        vqip["volume"] = excess
        return vqip

    def excess_to_garden_demand(self, excess):
        """Apply garden_efficiency.

        Args:
            excess (float): Volume of water required to satisfy garden irrigation

        Returns:
            (float): Amount of water actually applied to garden
        """
        # TODO Anything more than this needed? (yes - population presence if eventually
        # included!)

        return excess * self.gardening_efficiency

    def get_house_demand(self):
        """Per capita calculations for household wastewater generation. Applies weighted
        temperature calculation.

        Returns:
            (dict): A VQIP containg foul water
        """
        # TODO water that is consumed but not sent onwards as foul Total water required
        consumption = self.population * self.per_capita
        # Apply pollutants
        foul = self.copy_vqip(self.pollutant_load)
        # Scale to population
        for pol in constants.ADDITIVE_POLLUTANTS:
            foul[pol] *= self.population
        # Update volume and temperature (which is weighted based on air temperature and
        # constant_temp)
        foul["volume"] = consumption
        foul["temperature"] = (
            self.get_data_input("temperature") * (1 - self.constant_weighting)
            + self.constant_temp * self.constant_weighting
        )
        return foul



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson Converted to totals on 2022-05-03
"""
import warnings
from math import exp
from typing import Any, Dict

from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.tanks import DecayQueueTank, DecayTank, QueueTank, Tank


class Storage(Node):
    """"""

    def __init__(
        self,
        name,
        capacity=0,
        area=0,
        datum=0,
        decays=None,
        initial_storage=0,
        **kwargs,
    ):
        """A Node wrapper for a Tank or DecayTank.

        Args:
            name (str): node name capacity (float, optional): Tank capacity (see
            nodes.py/Tank). Defaults to 0. area (float, optional): Tank area (see
            nodes.py/Tank). Defaults to 0. datum (float, optional): Tank datum (see
            nodes.py/Tank). Defaults to 0. decays (dict, optional): Tank decays if
            needed, (see nodes.py/DecayTank). Defaults to None. initial_storage (float
            or dict, optional): Initial storage (see nodes.py/Tank). Defaults to 0.

        Functions intended to call in orchestration:
            distribute (optional, depends on subclass)
        """
        # Set parameters
        self.initial_storage = initial_storage
        self.capacity = capacity
        self.area = area
        self.datum = datum
        self.decays = decays
        super().__init__(name, **kwargs)

        # Create tank
        if "initial_storage" not in dir(self):
            self.initial_storage = self.empty_vqip()

        if self.decays is None:
            self.tank = Tank(
                capacity=self.capacity,
                area=self.area,
                datum=self.datum,
                initial_storage=self.initial_storage,
            )
        else:
            self.tank = DecayTank(
                capacity=self.capacity,
                area=self.area,
                datum=self.datum,
                initial_storage=self.initial_storage,
                decays=self.decays,
                parent=self,
            )
        # Update handlers
        self.push_set_handler["default"] = self.push_set_storage
        self.push_check_handler["default"] = self.tank.get_excess
        self.pull_set_handler["default"] = lambda vol: self.tank.pull_storage(vol)
        self.pull_check_handler["default"] = self.tank.get_avail

        # Mass balance
        self.mass_balance_ds.append(lambda: self.tank.ds())

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        capacity, area, datum, decays.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        # not using pop as these items need to stay
        # in the overrides to be fed into the tank overrides
        if "capacity" in overrides.keys():
            self.capacity = overrides["capacity"]
        if "area" in overrides.keys():
            self.area = overrides["area"]
        if "datum" in overrides.keys():
            self.datum = overrides["datum"]
        if "decays" in overrides.keys():
            if self.decays is None:
                raise ValueError(
                    "Attempting to override decays on a node initialised without decays"
                )
            self.decays.update(overrides["decays"])
        # apply tank overrides
        self.tank.apply_overrides(overrides)
        super().apply_overrides(overrides)

    def push_set_storage(self, vqip):
        """A node wrapper for the tank push_storage.

        Args:
            vqip (dict): A VQIP amount to push to the tank

        Returns:
            reply (dict): A VQIP amount that was not successfully pushed
        """
        # Update tank
        reply = self.tank.push_storage(vqip)

        return reply

    def distribute(self):
        """Optional function that discharges all tank storage with push_distributed."""
        # Distribute any active storage
        storage = self.tank.pull_storage(self.tank.get_avail())
        retained = self.push_distributed(storage)
        _ = self.tank.push_storage(retained, force=True)
        if retained["volume"] > constants.FLOAT_ACCURACY:
            print("Storage unable to push")

    def get_percent(self):
        """Function that returns the volume in the storage tank expressed as a percent
        of capacity."""
        return self.tank.storage["volume"] / self.tank.capacity

    def end_timestep(self):
        """Update tank states."""
        self.tank.end_timestep()

    def reinit(self):
        """Call tank reinit."""
        # TODO Automate this better
        self.tank.reinit()
        self.tank.storage["volume"] = self.initial_storage
        self.tank.storage_["volume"] = self.initial_storage


class Groundwater(Storage):
    """"""

    def __init__(
        self,
        residence_time=200,
        infiltration_threshold=1,
        infiltration_pct=0,
        data_input_dict={},
        **kwargs,
    ):
        # TODO why isn't this using a ResidenceTank?
        """A storage with a residence time for groundwater. Can also infiltrate to
        sewers.

        Args:
            residence_time (float, optional): Residence time (see nodes.py/
                ResidenceTank). Defaults to 200.
            infiltration_threshold (float, optional): Proportion of storage capacity
                that must be exceeded to generate infiltration. Defaults to 1.
            infiltration_pct (float, optional): Proportion of storage above the
                threshold that is square rooted and infiltrated. Defaults to 0.
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node (though I don't think it is used). Defaults to {}.

        Functions intended to call in orchestration:
            infiltrate (before sewers are discharged)

            distribute

        Key assumptions:
            - Conceptualises groundwater as a tank.
            - Baseflow is generated following a residence-time method.
            - Baseflow is sent to `storage.py/River`, `nodes.py/Node` or
                `waste.py/Waste` nodes.
            - Infiltration to `sewer.py/Sewer` nodes occurs when the storage
                in the tank is greater than a specified threshold, at a rate
                proportional to the sqrt of volume above the threshold. (Note, this
                behaviour is __not validated__ and a high uncertainty process in
                general)
            - If `decays` are provided to model water quality transformations,
                see `core.py/DecayObj`.

        Input data and parameter requirements:
            - Groundwater tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
            - Infiltration behaviour determined by an `infiltration_threshold`
                and `infiltration_pct`. _Units_: proportion of capacity
            - Optional dictionary of decays with pollutants as keys and decay
                parameters (a constant and a temperature sensitivity exponent) as
                values. _Units_: -
        """
        self.residence_time = residence_time
        self.infiltration_threshold = infiltration_threshold
        self.infiltration_pct = infiltration_pct
        # TODO not used data_input
        self.data_input_dict = data_input_dict
        super().__init__(**kwargs)

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        residence_time, infiltration_threshold, infiltration_pct.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.residence_time = overrides.pop("residence_time", self.residence_time)
        self.infiltration_threshold = overrides.pop(
            "infiltration_threshold", self.infiltration_threshold
        )
        self.infiltration_pct = overrides.pop("infiltration_pct", self.infiltration_pct)
        super().apply_overrides(overrides)

    def distribute(self):
        """Calculate outflow with residence time and send to Nodes or Rivers."""
        avail = self.tank.get_avail()["volume"] / self.residence_time
        to_send = self.tank.pull_storage({"volume": avail})
        retained = self.push_distributed(to_send, of_type=["Node", "River", "Waste"])
        _ = self.tank.push_storage(retained, force=True)
        if retained["volume"] > constants.FLOAT_ACCURACY:
            print("Storage unable to push")

    def infiltrate(self):
        """Calculate amount of water available for infiltration and send to sewers."""
        # Calculate infiltration
        avail = self.tank.get_avail()["volume"]
        avail = max(avail - self.tank.capacity * self.infiltration_threshold, 0)
        avail = (avail * self.infiltration_pct) ** 0.5

        # Push to sewers
        to_send = self.tank.pull_storage({"volume": avail})
        retained = self.push_distributed(to_send, of_type="Sewer")
        _ = self.tank.push_storage(retained, force=True)
        # Any not sent is left in tank
        if retained["volume"] > constants.FLOAT_ACCURACY:
            # print('unable to infiltrate')
            pass


class QueueGroundwater(Storage):
    """"""

    # TODO - no infiltration as yet
    def __init__(self, timearea={0: 1}, data_input_dict={}, **kwargs):
        """Alternate formulation of Groundwater that uses a timearea property to enable
        more nonlinear time behaviour of baseflow routing. Uses the QueueTank or
        DecayQueueTank (see nodes.py/Tank subclassses).

        NOTE: abstraction behaviour from this kind of node need careful checking

        Args:
            timearea (dict, optional): Time area diagram that enables flows to
                take a range of different durations to 'traverse' the tank. The keys of
                the dict are the number of timesteps while the values are the proportion
                of flow. E.g., {0 : 0.7, 1 : 0.3} means 70% of flow takes 0 timesteps
                and 30% takes 1 timesteps. Defaults to {0 : 1}.
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node (though I don't think it is used). Defaults to {}.

        Functions intended to call in orchestration:
            distribute

        Key assumptions:
            - Conceptualises groundwater as a tank.
            - Baseflow is generated following a timearea method.
            - Baseflow is sent to `storage.py/River`, `nodes.py/Node` or
                `waste.py/Waste` nodes.
            - No infiltration to sewers is modelled.
            - If `decays` are provided to model water quality transformations,
                see `core.py/DecayObj`.

        Input data and parameter requirements:
            - Groundwater tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
            - `timearea` is a dictionary containing the timearea diagram.
                _Units_: duration of flow (in timesteps) and proportion of flow
            - Optional dictionary of decays with pollutants as keys and decay
                parameters (a constant and a temperature sensitivity exponent) as
                values. _Units_: -
        """
        self.timearea = timearea
        # TODO not used
        self.data_input_dict = data_input_dict
        super().__init__(**kwargs)
        # Label as Groundwater class so that other nodes treat it the same
        self.__class__.__name__ = "Groundwater"
        # Update handlers
        self.push_set_handler["default"] = self.push_set_timearea
        self.pull_set_handler["default"] = self.pull_set_active
        self.pull_check_handler["default"] = self.pull_check_active
        # Enable decay
        if self.decays is None:
            self.tank = QueueTank(
                capacity=self.capacity,
                area=self.area,
                datum=self.datum,
                initial_storage=self.initial_storage,
            )
        else:
            self.tank = DecayQueueTank(
                capacity=self.capacity,
                area=self.area,
                datum=self.datum,
                decays=self.decays,
                parent=self,
                initial_storage=self.initial_storage,
            )

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        timearea.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.timearea = overrides.pop("timearea", self.timearea)
        super().apply_overrides(overrides)

    def push_set_timearea(self, vqip):
        """Push setting that enables timearea behaviour, (see __init__ for
        description).Used to receive flow that is assumed to occur widely across some
        kind of catchment.

        Args:
            vqip (dict): A VQIP that has been pushed

        Returns:
            reply (dict): A VQIP amount that was not successfuly receivesd
        """
        reply = self.empty_vqip()
        # Iterate over timearea diagram TODO timearea diagram behaviour be generalised
        # across nodes
        for time, normalised in self.timearea.items():
            vqip_ = self.v_change_vqip(vqip, vqip["volume"] * normalised)
            reply_ = self.tank.push_storage(vqip_, time=time)
            reply = self.sum_vqip(reply, reply_)
        return reply

    def distribute(self):
        """Update internal arc, push active_storage onwards, update tank."""
        _ = self.tank.internal_arc.update_queue(direction="push")

        remaining = self.push_distributed(self.tank.active_storage)

        if remaining["volume"] > constants.FLOAT_ACCURACY:
            print("Groundwater couldnt push all")

        # Update tank
        sent = self.tank.active_storage["volume"] - remaining["volume"]
        sent = self.v_change_vqip(self.tank.active_storage, sent)
        reply = self.tank.pull_storage(sent)
        if (reply["volume"] - sent["volume"]) > constants.FLOAT_ACCURACY:
            print("Miscalculated tank storage in discharge")

    def infiltrate(self):
        """"""
        pass

    def pull_check_active(self, vqip=None):
        """A pull check that returns the active storage.

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            (dict): A VQIP amount that is available to pull
        """
        if vqip is None:
            return self.tank.active_storage
        else:
            reply = min(vqip["volume"], self.tank.active_storage["volume"])
            return self.v_change_vqip(self.tank.active_storage, reply)

    def pull_set_active(self, vqip):
        # TODO - this is quite weird behaviour, and inconsistent with pull_check_active
        """Pull proportionately from both the active storage and the queue. Adjudging
        groundwater abstractions to not be particularly sensitive to the within
        catchment travel time.

        Args:
            vqip (dict): A VQIP amount to be pulled (only volume key is used)

        Returns:
            pulled (dict): A VQIP amount that was successfully pulled
        """
        # Calculate actual pull
        total_storage = self.tank.storage["volume"]
        total_pull = min(self.tank.storage["volume"], vqip["volume"])

        if total_pull < constants.FLOAT_ACCURACY:
            return self.empty_vqip()
        else:
            # Track total pull in pulled
            pulled = self.empty_vqip()
            # Iterate over queue
            if isinstance(self.tank.internal_arc.queue, dict):
                for t, v in self.tank.internal_arc.queue.items():
                    # Pull proportionately
                    t_pulled = self.v_change_vqip(
                        self.tank.internal_arc.queue[t],
                        v["volume"] * total_pull / total_storage,
                    )
                    # Reduce queue VQIPs
                    self.tank.internal_arc.queue[t] = self.extract_vqip(
                        self.tank.internal_arc.queue[t], t_pulled
                    )
                    # Track pull
                    pulled = self.sum_vqip(pulled, t_pulled)
                # Pull also from active storage
                a_pulled = self.v_change_vqip(
                    self.tank.active_storage,
                    self.tank.active_storage["volume"] * total_pull / total_storage,
                )
                self.tank.active_storage = self.extract_vqip(
                    self.tank.active_storage, a_pulled
                )
                pulled = self.sum_vqip(pulled, a_pulled)

                # Recalculate storage
                self.tank.storage = self.extract_vqip(self.tank.storage, pulled)
                return pulled
            # elif isinstance(self.tank.internal_arc.queue, list): for req in
            #     self.tank.internal_arc.queue: t_pulled = req['vqtip']['volume'] *
            #         total_pull / total_storage req['vqtip'] =
            #         self.v_change_vqip(req['vqtip'], req['vqtip']['volume'] -
            #         t_pulled) pulled += t_pulled a_pulled =
            #     self.tank.active_storage['volume'] * total_pull / total_storage
            #     self.tank.active_storage =
            #     self.v_change_vqip(self.tank.active_storage,
            #     self.tank.active_storage['volume'] - a_pulled) pulled += a_pulled

            #     #Recalculate storage - doing this differently causes numerical errors
            #     new_v = sum([x['vqtip']['volume'] for x in
            #     self.tank.internal_arc.queue])+ self.tank.active_storage['volume']
            #     self.tank.storage = self.v_change_vqip(self.tank.storage, new_v)

            # return self.v_change_vqip(self.tank.storage, pulled)


class River(Storage):
    """"""

    # TODO non-day timestep
    def __init__(
        self,
        depth=2,
        length=200,
        width=20,
        velocity=0.2 * constants.M_S_TO_M_DT,
        damp=0.1,
        mrf=0,
        **kwargs,
    ):
        """Node that contains extensive in-river biochemical processes.

        Args:
            depth (float, optional): River tank depth. Defaults to 2. length (float,
            optional): River tank length. Defaults to 200. width (float, optional):
            River tank width. Defaults to 20. velocity (float, optional): River velocity
            (if someone wants to calculate
                this on the fly that would also work). Defaults to
                0.2*constants.M_S_TO_M_DT.
            damp (float, optional): Flow delay and attentuation parameter. Defaults
                to 0.1.
            mrf (float, optional): Minimum required flow in river (volume per timestep),
                can limit pulls made to the river. Defaults to 0.

        Functions intended to call in orchestration:
            distribute

        Key assumptions:
             - River is conceptualised as a water tank that receives flows from various
                sources (e.g., runoffs from urban and rural land, baseflow from
                groundwater), interacts with water infrastructure (e.g., abstraction for
                irrigation and domestic supply, sewage and treated effluent discharge),
                and discharges flows downstream. It has length and width as shape
                parameters, average velocity to indicate flow speed and capacity to
                indicate the maximum storage limit.
             - Flows from different sources into rivers will fully mix. River tank is
               assumed to
                have delay and attenuation effects when generate outflows. These effects
                are simulated based on the average velocity.
             - In-river biochemical processes are simulated as sources/sinks of
               nutrients
                in the river tank, including - denitrification (for nitrogen) -
                phytoplankton absorption/release (for nitrogen and phosphorus) -
                macrophyte uptake (for nitrogen and phosphorus) These processes are
                affected by river temperature.

        Input data and parameter requirements:
             - depth, length, width
                _Units_: m
             - velocity
                _Units_: m/day
             - damping coefficient
                _Units_: -
             - minimum required flow
                _Units_: m3/day
        """
        # Set parameters
        self.depth = depth
        if depth != 2:
            warnings.warn(
                "warning: the depth parameter is unused by River nodes because it is \
		intended for capacity to be unbounded. It may be removed in a future version."
            )
        self.length = length  # [m]
        self.width = width  # [m]
        self.velocity = velocity  # [m/dt]
        self.damp = damp  # [>=0] flow delay and attenuation
        self.mrf = mrf
        area = length * width  # [m2]

        capacity = (
            constants.UNBOUNDED_CAPACITY
        )  # TODO might be depth * area if flood indunation is going to be simulated

        # Required in cases where 'area' conflicts with length*width
        kwargs["area"] = area
        # Required in cases where 'capacity' conflicts with depth*area
        kwargs["capacity"] = capacity

        super().__init__(**kwargs)

        # TODO check units TODO Will a user want to change any of these? Wide variety of
        # river parameters (from HYPE)
        self.uptake_PNratio = 1 / 7.2  # [-] P:N during crop uptake
        self.bulk_density = 1300  # [kg/m3] soil density
        self.denpar_w = 0.0015  # 0.001, # [kg/m2/day] reference denitrification rate
        # in water course
        self.T_wdays = 5  # [days] weighting constant for river temperature calculation
        # (similar to moving average period)
        self.halfsatINwater = (
            1.5 * constants.MG_L_TO_KG_M3
        )  # [kg/m3] half saturation parameter for denitrification in river
        self.T_10_days = []  # [degree C] average water temperature of 10 days
        self.T_20_days = []  # [degree C] average water temperature of 20 days
        self.TP_365_days = []  # [degree C] average water temperature of 20 days
        self.hsatTP = 0.05 * constants.MG_L_TO_KG_M3  # [kg/m3]
        self.limpppar = 0.1 * constants.MG_L_TO_KG_M3  # [kg/m3]
        self.prodNpar = 0.001  # [kg N/m3/day] nitrogen production/mineralisation rate
        self.prodPpar = (
            0.0001  # [kg N/m3/day] phosphorus production/mineralisation rate
        )
        self.muptNpar = 0.001  # [kg/m2/day] nitrogen macrophyte uptake rate
        self.muptPpar = 0.0001  # 0.01, # [kg/m2/day] phosphorus macrophyte uptake rate

        self.max_temp_lag = 20
        self.lagged_temperatures = []

        self.max_phosphorus_lag = 365
        self.lagged_total_phosphorus = []

        self.din_components = ["ammonia", "nitrate"]
        # TODO need a cleaner way to do this depending on whether e.g., nitrite is
        # included

        # Initialise paramters
        self.current_depth = 0  # [m]
        # self.river_temperature = 0 # [degree C] self.river_denitrification = 0 #
        # [kg/day] self.macrophyte_uptake_N = 0 # [kg/day] self.macrophyte_uptake_P = 0
        # # [kg/day] self.sediment_particulate_phosphorus_pool = 60000 # [kg]
        # self.sediment_pool = 1000000 # [kg] self.benthos_source_sink = 0 # [kg/day]
        # self.t_res = 0 # [day] self.outflow = self.empty_vqip()

        # Update end_teimstep
        self.end_timestep = self.end_timestep_

        # Update handlers
        self.push_set_handler["default"] = self.push_set_river
        self.push_check_handler["default"] = self.push_check_accept

        self.pull_check_handler["default"] = self.pull_check_river
        self.pull_set_handler["default"] = self.pull_set_river

        # TODO - RiparianBuffer
        self.pull_check_handler[("RiparianBuffer", "volume")] = self.pull_check_fp

        # Update mass balance
        self.bio_in = self.empty_vqip()
        self.bio_out = self.empty_vqip()

        self.mass_balance_in.append(lambda: self.bio_in)
        self.mass_balance_out.append(lambda: self.bio_out)

    # TODO something like this might be needed if you want sewers backing up from river
    # height... would need to incorporate expected river outflow def get_dt_excess(self,
    #     vqip = None): reply = self.empty_vqip() reply['volume'] =
    #     self.tank.get_excess()['volume'] + self.tank.get_avail()['volume'] *
    #     self.get_riverrc() if vqip is not None: reply['volume'] = min(vqip['volume'],
    #         reply['volume']) return reply

    # def push_set_river(self, vqip): vqip_ = vqip.copy() vqip_ =
    #     self.v_change_vqip(vqip_, min(vqip_['volume'],
    #     self.get_dt_excess()['volume'])) _ = self.tank.push_storage(vqip_, force=True)
    #     return self.extract_vqip(vqip, vqip_)

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        timearea.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        overwrite_params = set(
            [
                "length",
                "width",
                "velocity",
                "damp",
                "mrf",
                "uptake_PNratio",
                "bulk_density",
                "denpar_w",
                "T_wdays",
                "halfsatINwater",
                "hsatTP",
                "limpppar",
                "prodNpar",
                "prodPpar",
                "muptNpar",
                "muptPpar",
                "max_temp_lag",
                "max_phosphorus_lag",
            ]
        )

        for param in overwrite_params.intersection(overrides.keys()):
            setattr(self, param, overrides.pop(param))

        if "area" in overrides.keys():
            warnings.warn(
                "WARNING: specifying area is depreciated in overrides \
		for river, please specify length and width instead"
            )
        overrides["area"] = self.length * self.width
        if "capacity" in overrides.keys():
            warnings.warn(
                "ERROR: specifying capacity is depreciated in overrides \
		for river, it is always set as unbounded capacity"
            )
        overrides["capacity"] = constants.UNBOUNDED_CAPACITY
        super().apply_overrides(overrides)

    def pull_check_river(self, vqip=None):
        """Check amount of water that can be pulled from river tank and upstream.

        Args:
            vqip (dict, optional): Maximum water required (only 'volume' is used)

        Returns:
            avail (dict): A VQIP amount that can be pulled
        """
        # Get storage
        avail = self.tank.get_avail()

        # Check incoming
        upstream = self.get_connected(direction="pull", of_type=["River", "Node"])
        avail["volume"] += upstream["avail"]

        # convert mrf from volume/timestep to discrete value
        mrf = self.mrf / self.get_riverrc()

        # Apply mrf
        avail_vol = max(avail["volume"] - mrf, 0)
        if vqip is None:
            avail = self.v_change_vqip(avail, avail_vol)
        else:
            avail = self.v_change_vqip(avail, min(avail_vol, vqip["volume"]))

        return avail

    def pull_set_river(self, vqip):
        """Pull from river tank and upstream, acknowledging MRF with pull_check.

        Args:
            vqip (dict): A VQIP amount to pull (only volume key used)

        Returns:
            (dict): A VQIP amount that was pulled
        """
        # Calculate available pull
        avail = self.pull_check_river(vqip)

        # Take first from tank
        pulled = self.tank.pull_storage(avail)

        # Take remaining from upstream
        to_pull = {"volume": avail["volume"] - pulled["volume"]}
        pulled_ = self.pull_distributed(to_pull, of_type=["River", "Node"])

        reply = self.sum_vqip(pulled, pulled_)

        return reply

    def push_set_river(self, vqip):
        """Push to river tank, currently forced.

        Args:
            vqip (dict): A VQIP amount to push

        Returns:
            (dict): A VQIP amount that was not successfully received
        """
        _ = self.tank.push_storage(vqip, force=True)
        return self.empty_vqip()

    def update_depth(self):
        """Recalculate depth."""
        self.current_depth = self.tank.storage["volume"] / self.area

    def get_din_pool(self):
        """Get total dissolved inorganic nitrogen from tank storage.

        Returns:
            (float): total din
        """
        return sum(
            [self.tank.storage[x] for x in self.din_components]
        )  # TODO + self.tank.storage['nitrite'] but nitrite might not be modelled...
        # need some ways to address this

    def biochemical_processes(self):
        """Runs all biochemical processes and updates pollutant amounts.

        Returns:
            in_ (dict): A VQIP amount that represents total gain in pollutant amounts
            out_ (dict): A VQIP amount that represents total loss in pollutant amounts
        """
        # TODO make more modular
        self.update_depth()

        self.tank.storage["temperature"] = (1 - 1 / self.T_wdays) * self.tank.storage[
            "temperature"
        ] + (1 / self.T_wdays) * self.get_data_input("temperature")

        # Update lagged temperatures
        if len(self.lagged_temperatures) > self.max_temp_lag:
            del self.lagged_temperatures[0]
        self.lagged_temperatures.append(self.tank.storage["temperature"])

        # Update lagged total phosphorus
        if len(self.lagged_total_phosphorus) > self.max_phosphorus_lag:
            del self.lagged_total_phosphorus[0]
        total_phosphorus = (
            self.tank.storage["phosphate"] + self.tank.storage["org-phosphorus"]
        )
        self.lagged_total_phosphorus.append(total_phosphorus)

        # Check if any water
        if self.tank.storage["volume"] < constants.FLOAT_ACCURACY:
            # Assume these only do something when there is water
            return (self.empty_vqip(), self.empty_vqip())

        if self.tank.storage["temperature"] <= 0:
            # Seems that these things are only active when above freezing
            return (self.empty_vqip(), self.empty_vqip())

        # Denitrification
        tempfcn = 2 ** ((self.tank.storage["temperature"] - 20) / 10)
        if self.tank.storage["temperature"] < 5:
            tempfcn *= self.tank.storage["temperature"] / 5

        din = self.get_din_pool()
        din_concentration = din / self.tank.storage["volume"]
        confcn = din_concentration / (
            din_concentration + self.halfsatINwater
        )  # [kg/m3]
        denitri_water = (
            self.denpar_w * self.area * tempfcn * confcn
        )  # [kg/day] #TODO convert to per DT

        river_denitrification = min(
            denitri_water, 0.5 * din
        )  # [kg/day] max 50% kan be denitrified
        din_ = din - river_denitrification  # [kg]

        # Update mass balance
        in_ = self.empty_vqip()
        out_ = self.empty_vqip()
        if din > 0:
            for pol in self.din_components:
                # denitrification
                loss = (din - din_) / din * self.tank.storage[pol]
                out_[pol] += loss
                self.tank.storage[pol] -= loss

        din = self.get_din_pool()

        # Calculate moving averages TODO generalise
        temp_10_day = sum(self.lagged_temperatures[-10:]) / 10
        temp_20_day = sum(self.lagged_temperatures[-20:]) / 20
        total_phos_365_day = sum(self.lagged_total_phosphorus) / self.max_phosphorus_lag

        # Calculate coefficients
        tempfcn = (
            (self.tank.storage["temperature"]) / 20 * (temp_10_day - temp_20_day) / 5
        )
        if (total_phos_365_day - self.limpppar + self.hsatTP) > 0:
            totalphosfcn = (total_phos_365_day - self.limpppar) / (
                total_phos_365_day - self.limpppar + self.hsatTP
            )
        else:
            totalphosfcn = 0

        # Mineralisation/production TODO this feels like it could be much tidier
        minprodN = (
            self.prodNpar * totalphosfcn * tempfcn * self.area * self.current_depth
        )  # [kg N/day]
        minprodP = (
            self.prodPpar
            * totalphosfcn
            * tempfcn
            * self.area
            * self.current_depth
            * self.uptake_PNratio
        )  # [kg N/day]
        if minprodN > 0:
            # production (inorg -> org)
            minprodN = min(
                0.5 * din, minprodN
            )  # only half pool can be used for production
            minprodP = min(
                0.5 * self.tank.storage["phosphate"], minprodP
            )  # only half pool can be used for production

            # Update mass balance
            out_["phosphate"] = minprodP
            self.tank.storage["phosphate"] -= minprodP
            in_["org-phosphorus"] = minprodP
            self.tank.storage["org-phosphorus"] += minprodP
            if din > 0:
                for pol in self.din_components:
                    loss = minprodN * self.tank.storage[pol] / din
                    out_[pol] += loss
                    self.tank.storage[pol] -= loss

            in_["org-nitrogen"] = minprodN
            self.tank.storage["org-nitrogen"] += minprodN

        else:
            # mineralisation (org -> inorg)
            minprodN = min(0.5 * self.tank.storage["org-nitrogen"], -minprodN)
            minprodP = min(0.5 * self.tank.storage["org-phosphorus"], -minprodP)

            # Update mass balance
            in_["phosphate"] = minprodP
            self.tank.storage["phosphate"] += minprodP
            out_["org-phosphorus"] = minprodP
            self.tank.storage["org-phosphorus"] -= minprodP
            if din > 0:
                for pol in self.din_components:
                    gain = minprodN * self.tank.storage[pol] / din
                    in_[pol] += gain
                    self.tank.storage[pol] += gain

            out_["org-nitrogen"] = minprodN
            self.tank.storage["org-nitrogen"] -= minprodN

        din = self.get_din_pool()

        # macrophyte uptake temperature dependence factor
        tempfcn1 = (max(0, self.tank.storage["temperature"]) / 20) ** 0.3
        tempfcn2 = (self.tank.storage["temperature"] - temp_20_day) / 5
        tempfcn = max(0, tempfcn1 * tempfcn2)

        macrouptN = self.muptNpar * tempfcn * self.area  # [kg/day]
        macrophyte_uptake_N = min(0.5 * din, macrouptN)
        if din > 0:
            for pol in self.din_components:
                loss = macrophyte_uptake_N * self.tank.storage[pol] / din
                out_[pol] += loss
                self.tank.storage[pol] -= loss

        macrouptP = (
            self.muptPpar * tempfcn * max(0, totalphosfcn) * self.area
        )  # [kg/day]
        macrophyte_uptake_P = min(0.5 * self.tank.storage["phosphate"], macrouptP)
        out_["phosphate"] += macrophyte_uptake_P
        self.tank.storage["phosphate"] -= macrophyte_uptake_P

        # TODO source/sink for benthos sediment P suspension/resuspension
        return in_, out_

    def get_riverrc(self):
        """Get river outflow coefficient (i.e., how much water leaves the tank in this
        timestep).

        Returns:
            riverrc (float): outflow coeffficient
        """
        # Calculate travel time
        total_time = self.length / self.velocity
        # Apply damp
        kt = self.damp * total_time  # [day]
        if kt != 0:
            riverrc = 1 - kt + kt * exp(-1 / kt)  # [-]
        else:
            riverrc = 1
        return riverrc

    def calculate_discharge(self):
        """"""
        if "nitrate" in constants.POLLUTANTS:
            # TODO clumsy Run biochemical processes
            in_, out_ = self.biochemical_processes()
            # Mass balance
            self.bio_in = in_
            self.bio_out = out_

    def distribute(self):
        """Run biochemical processes, track mass balance, and distribute water."""
        # self.calculate_discharge() Get outflow
        outflow = self.tank.pull_storage(
            {"volume": self.tank.storage["volume"] * self.get_riverrc()}
        )
        # Distribute outflow
        reply = self.push_distributed(outflow, of_type=["River", "Node", "Waste"])
        _ = self.tank.push_storage(reply, force=True)
        if reply["volume"] > constants.FLOAT_ACCURACY:
            print("river cant push: {0}".format(reply["volume"]))

    def pull_check_fp(self, vqip=None):
        """

        Args:
            vqip:

        Returns:

        """
        # TODO Pull checking for riparian buffer, needs updating update river depth
        self.update_depth()
        return self.current_depth, self.area, self.width, self.river_tank.storage

    def end_timestep_(self):
        """Update state variables."""
        self.tank.end_timestep()
        self.bio_in = self.empty_vqip()
        self.bio_out = self.empty_vqip()


class Reservoir(Storage):
    """"""

    def __init__(self, **kwargs):
        """Storage node that makes abstractions by calling pull_distributed.

        Functions intended to call in orchestration:
            make_abstractions (before any river routing)

        Key assumptions:
            - Conceptualised as a `Tank`.
            - Recharged only via pumped abstractions.
            - Evaporation/precipitation onto surface area currently ignored.
            - If `decays` are provided to model water quality transformations,
                see `core.py/DecayObj`.

        Input data and parameter requirements:
            - Tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
            - Optional dictionary of decays with pollutants as keys and decay
                parameters (a constant and a temperature sensitivity exponent) as
                values. _Units_: -
        """
        super().__init__(**kwargs)

    def make_abstractions(self):
        """Pulls water and updates tanks."""
        reply = self.pull_distributed(self.tank.get_excess())
        spill = self.tank.push_storage(reply)
        _ = self.tank.push_storage(spill, force=True)
        if spill["volume"] > constants.FLOAT_ACCURACY:
            print("Spill at reservoir by {0}".format(spill["volume"]))


class RiverReservoir(Reservoir):
    """"""

    def __init__(self, environmental_flow=0, **kwargs):
        """A reservoir with a natural river inflow, includes an environmental downstream
        flow to satisfy.

        Args:
            environmental_flow (float, optional): Downstream environmental flow.
                Defaults to 0.

        Functions intended to call in orchestration:
                make_abstractions (if any)

                satisfy_environmental (before river routing.. possibly before
                    downstream river abstractions depending on licence)

        Key assumptions:
            - Conceptualised as a `Tank`.
            - Recharged via pumped abstractions and receives water from
                inflowing arcs.
            - Reservoir aims to satisfy a static `environmental_flow`.
            - If tank capacity is exceeded, reservoir spills downstream
                towards `nodes.py/Node`, `storage.py/River` or `waste.py/Waste` nodes.
                Spill counts towards `environmental_flow`.
            - Evaporation/precipitation onto surface area currently ignored.
            - Currently, if a reservoir satisfies a pull from a downstream
                node, it does __not__ count towards `environmental_flow`.
            - If `decays` are provided to model water quality transformations,
                see `core.py/DecayObj`.

        Input data and parameter requirements:
            - Tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
            - `environmental_flow`
                _Units_: cubic metres/timestep
            - Optional dictionary of decays with pollutants as keys and decay
                parameters (a constant and a temperature sensitivity exponent) as
                values. _Units_: -
        """
        # Parameters
        self.environmental_flow = environmental_flow
        super().__init__(**kwargs)

        # State variables
        self.total_environmental_satisfied = 0

        self.push_set_handler["default"] = self.push_set_river_reservoir
        self.push_check_handler["default"] = self.push_check_river_reservoir
        self.end_timestep = self.end_timestep_

        self.__class__.__name__ = "Reservoir"

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        environmental_flow.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.environmental_flow = overrides.pop(
            "environmental_flow", self.environmental_flow
        )
        super().apply_overrides(overrides)

    def push_set_river_reservoir(self, vqip):
        """Receive water.

        Args:
            vqip (dict): A VQIP amount to be received

        Returns:
            reply (dict): A VQIP amount that was not successfully received
        """
        # Apply normal reservoir storage We do this under the assumption that spill is
        # mixed in with the reservoir If the reservoir can't spill everything you'll get
        # some weird numbers in reply, but if your reservoir can't spill as much as you
        # like then you should probably be pushing the right amount through
        # push_check_river_reservoir Some cunning could probably avoid this by checking
        # vqip, but this is a serious edge case
        _ = self.tank.push_storage(vqip, force=True)
        spill = self.tank.pull_ponded()

        # Send spill downstream
        reply = self.push_distributed(spill, of_type=["Node", "River", "Waste"])

        # Use spill to satisfy downstream flow
        self.total_environmental_satisfied += spill["volume"] - reply["volume"]

        return reply

    def push_check_river_reservoir(self, vqip=None):
        """A push check to receive water, assumes spill may occur and checks downstream
        capacity.

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            excess (dict): A VQIP amount of water that cannot be received
        """
        # Check downstream capacity (i.e., that would be spilled)
        downstream_availability = self.get_connected(
            direction="push", of_type=["Node", "River", "Waste"]
        )["avail"]
        # Check excess capacity in the reservoir
        excess = self.tank.get_excess()
        # Combine excess and downstream in response
        new_v = excess["volume"] + downstream_availability
        if vqip is not None:
            new_v = min(vqip["volume"], new_v)
        # Update to vqip
        excess = self.v_change_vqip(excess, new_v)

        return excess

    def satisfy_environmental(self):
        """Satisfy environmental flow downstream."""
        # Calculate how much environmental flow is yet to satisfy #(some may have been
        # already if pull-and-take abstractions have taken place)
        to_satisfy = max(
            self.environmental_flow - self.total_environmental_satisfied, 0
        )
        # Pull from tank
        environmental = self.tank.pull_storage({"volume": to_satisfy})
        # Send downstream
        reply = self.push_distributed(environmental)
        _ = self.tank.push_storage(reply, force=True)
        if reply["volume"] > constants.FLOAT_ACCURACY:
            print("warning: environmental not able to push")

        # Update satisfaction
        self.total_environmental_satisfied += environmental["volume"]

    def end_timestep_(self):
        """Udpate state varibles."""
        self.tank.end_timestep()
        self.total_environmental_satisfied = 0



# %% [markdown]
# :warning: Warning: this code does not represent best practice for
# customisation. Instead we recommend use of decorators to overwrite
# behaviour. Users can see examples of decorators in the
# [customise_interactions](https://barneydobson.github.io/wsi/demo/scripts/customise_interactions/)
# and [customise_riverreservoir](https://barneydobson.github.io/wsi/demo/scripts/customise_riverreservoir/)
# guides. This guide may still be useful for WSIMOD code examples, and it
# will still work in many cases.
#
# There is currently a GitHub (issue)[https://github.com/barneydobson/wsi/issues/11]
# to revise this guide.
#
# # Customise an arc (.py)
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/barneydobson/wsi/blob/main/docs/demo/scripts/customise_an_arc.py)
#
# 1. [Introduction](#introduction)
#
# 2. [Create baseline](#create-baseline)
#
# 3. [Customise arc](#customise-arc)
#
# 4. [Inspect results](#inspect-results)
#
# %% [markdown]
# ## Introduction
# In this tutorial we will demonstrate how to customise an arc on-the-fly. If
# you are creating a new type of arc that is a generic physical object (e.g., a
# weir), then you should probably create a new subclass of existing classes.
# However, if you are aiming to alter the behaviour of an arc that already has
# a class to fit some super specific behaviour, then it may be more suitable
# to customise on-the-fly. In addition, customising on-the-fly is very similar
# to creating a subclass, so the skills demonstrated here are likely to be
# transferrable.
#
# We will use the [Oxford demo](./../oxford_demo) as our base model, and build
# off it by assigning a varying capacity to the river abstraction depending on
# reservoir storage
#
# ## Create baseline
# We will first create and simulate the Oxford model to formulate baseline
# results. We use a version of the model with the minimum required flow from
# the [node customisation demo](./../customise_a_node).
#
# Start by importing packages.
# %%
import os

import pandas as pd
from matplotlib import pyplot as plt

from wsimod.demo.create_oxford import create_oxford_model_mrf

# %% [markdown]
# The model can be automatically created using the data_folder
# %%
# Select the root path for the data folder. Use the appropriate value for your case.
data_folder = os.path.join(os.path.abspath(""), "docs", "demo", "data")

baseline_model = create_oxford_model_mrf(data_folder)

# %% [markdown]
# Simulate baseline flows
# %%
baseline_flows, baseline_tanks, _, _ = baseline_model.run()
baseline_flows = pd.DataFrame(baseline_flows)
baseline_tanks = pd.DataFrame(baseline_tanks)

# %% [markdown]
# When we plot the results, we see some alarmingly low reservoir levels - I am sure that Thames Water
# would not be happy about this
# %%
f, axs = plt.subplots(3, 1, figsize=(6, 6))
baseline_flows.groupby("arc").get_group("farmoor_to_mixer").set_index("time")[
    ["flow"]
].plot(ax=axs[0], title="River flow downstream of abstractions")
axs[0].set_yscale("symlog")
axs[0].set_ylim([10e3, 10e7])
baseline_flows.groupby("arc").get_group("abstraction_to_farmoor").set_index("time")[
    ["flow"]
].plot(ax=axs[1], title="Abstraction")
axs[1].legend()

baseline_tanks.groupby("node").get_group("farmoor").set_index("time")[["storage"]].plot(
    ax=axs[2], title="Reservoir storage"
)
axs[2].legend()
f.tight_layout()

# %% [markdown]
# ## Customise arc
# The easiest way to customise an arc is by changing its ```get_excess```
# function. This is the function that determines how much capacity there is in
# the arc when called.
#
# In reality it is common for operational constraints to be seasonal, so we will
# define monthly amounts that, when reservoir levels are below them, the
# abstraction capacity can be increased. These amounts have the month as the
# key and the percentage that the reservoir is full as values.
# %%

levels = {
    1: 0.7,
    2: 0.8,
    3: 0.9,
    4: 0.95,
    5: 0.95,
    6: 0.95,
    7: 0.95,
    8: 0.9,
    9: 0.8,
    10: 0.7,
    11: 0.7,
    12: 0.7,
}
# %% [markdown]
# We also need to determine what the new abstraction capacity will be, we will
# choose a simple multiplier on the existing capacity
# %%
capacity_multiplier = 1.25


# %% [markdown]
# We can now redefine the ```get_excess``` function. We can start with a trimmed down
# version of the original ```get_excess``` function:
# ```
# def get_excess(self, direction, vqip = None, tag = 'default'):
#     pipe_excess = self.capacity - self.flow_in
#     node_excess = self.in_port.pull_check(vqip, tag)
#     excess = min(pipe_excess, node_excess['volume'])
#     return self.v_change_vqip(node_excess, excess)
# ```
# This shows only the bits of the function that are relevant for pulls.
# In order to implement the new variable capacity, we will need to:
#
#   -identify the month
#
#   -identify the reservoir volume expressed as a percent
#
#   -adjust the capacity to reflect possibility of increased abstractions
# %%
def get_excess_new(arc, direction, vqip=None, tag="default"):
    """

    Args:
        arc:
        direction:
        vqip:
        tag:

    Returns:

    """
    # All nodes have access to the 't' parameter, which is a datetime object
    # that can return the month
    month = arc.out_port.t.month

    # Get percent full
    pct = arc.out_port.get_percent()

    # Adjust capacity
    if arc.levels[month] > pct:
        capacity = arc.capacity * arc.capacity_multiplier
    else:
        capacity = arc.capacity

    # Proceed with get_excess function as normal
    pipe_excess = capacity - arc.flow_in
    node_excess = arc.in_port.pull_check(vqip, tag)
    excess = min(pipe_excess, node_excess["volume"])
    return arc.v_change_vqip(node_excess, excess)


# %% [markdown]
# Finally we need a wrapper to assign the new function and the associated
# parameters
# %%
def apply_variable_capacity(arc, levels, multiplier):
    """

    Args:
        arc:
        levels:
        multiplier:
    """
    # Assign parameters
    arc.levels = levels
    arc.capacity_multiplier = multiplier

    # Change get_excess function to new one above
    arc.get_excess = lambda **x: get_excess_new(arc, **x)


# %% [markdown]
# Now we can create a new model
# %%
customised_model = create_oxford_model_mrf(data_folder)

# %% [markdown]
# .. and assign the new variable capacity
# %%
apply_variable_capacity(
    customised_model.arcs["abstraction_to_farmoor"], levels, capacity_multiplier
)

# %% [markdown]
# ## Inspect results
# Let us rerun and view the results to see if it has worked.
# %%

flows_var_cap, tanks_var_cap, _, _ = customised_model.run()

flows_var_cap = pd.DataFrame(flows_var_cap)
tanks_var_cap = pd.DataFrame(tanks_var_cap)

# %% [markdown]
# We can see that the reservoir storage is able to recharge much more quickly
# from the lowest point because of the increased abstractions. We also see
# negligible change in river flow because the increased abstractions are mainly
# occurring when the flow is quite high. This is still unlikely to be a
# preferred option in practice because the lowest reservoir storage is
# unchanged. This happens because the increased capacity available is not helpful when the flows are
# so low that the minimum required flow is active. Better luck next time Thames Water!
# %%

f, axs = plt.subplots(3, 1, figsize=(6, 6))
plot_flows1 = pd.concat(
    [
        baseline_flows.groupby("arc")
        .get_group("farmoor_to_mixer")
        .set_index("time")
        .flow.rename("Baseline"),
        flows_var_cap.groupby("arc")
        .get_group("farmoor_to_mixer")
        .set_index("time")
        .flow.rename("Customised"),
    ],
    axis=1,
)
plot_flows1.plot(ax=axs[0], title="River flow downstream of abstractions")
axs[0].set_yscale("symlog")
axs[0].set_ylim([10e3, 10e7])

plot_flows2 = pd.concat(
    [
        baseline_flows.groupby("arc")
        .get_group("abstraction_to_farmoor")
        .set_index("time")
        .flow.rename("Baseline"),
        flows_var_cap.groupby("arc")
        .get_group("abstraction_to_farmoor")
        .set_index("time")
        .flow.rename("Customised"),
    ],
    axis=1,
)
plot_flows2.plot(ax=axs[1], title="Abstraction")
axs[1].legend()

plot_tanks = pd.concat(
    [
        baseline_tanks.groupby("node")
        .get_group("farmoor")
        .set_index("time")
        .storage.rename("Baseline"),
        tanks_var_cap.groupby("node")
        .get_group("farmoor")
        .set_index("time")
        .storage.rename("Customised"),
    ],
    axis=1,
)
plot_tanks.plot(ax=axs[2], title="Reservoir storage")
axs[2].legend()
f.tight_layout()



# ruff: noqa: F401
from wsimod.nodes.catchment import Catchment
from wsimod.nodes.demand import Demand, NonResidentialDemand, ResidentialDemand
from wsimod.nodes.distribution import Distribution, UnlimitedDistribution
from wsimod.nodes.land import Land
from wsimod.nodes.nodes import NODES_REGISTRY, Node
from wsimod.nodes.sewer import EnfieldFoulSewer, Sewer
from wsimod.nodes.storage import (
    Groundwater,
    QueueGroundwater,
    Reservoir,
    River,
    RiverReservoir,
    Storage,
)
from wsimod.nodes.waste import Waste
from wsimod.nodes.wtw import FWTW, WTW, WWTW



# -*- coding: utf-8 -*-
"""Created on Tue Nov  1 09:03:17 2022.

@author: Barney
"""
import os
import re

import pandas as pd

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.catchment import Catchment
from wsimod.nodes.demand import ResidentialDemand
from wsimod.nodes.land import Land
from wsimod.nodes.nodes import Node
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.storage import Groundwater, Reservoir
from wsimod.nodes.waste import Waste
from wsimod.nodes.wtw import FWTW, WWTW
from wsimod.orchestration.model import Model

# NOTE - these are only supplements to the demos in the documentation
# if you want to see demos, go to docs/demo/scripts!


def preprocess_oxford(data_dir=None):
    """

    Args:
        data_dir:
    """
    # This function is used to create the timeseries data for the oxford demo
    if data_dir is None:
        data_dir = os.path.join(
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            ),
            "demo",
            "data",
        )

    UG_L_TO_KG_M3 = 1e-6
    MG_L_TO_KG_M3 = 1e-3

    q_lab = [
        ("39008_gdf.csv", "thames"),
        ("39034_gdf.csv", "evenlode"),
        ("39021_gdf.csv", "cherwell"),
        ("39140_gdf.csv", "ray"),
    ]

    flows = []
    for fn, river in q_lab:
        df = pd.read_csv(
            os.path.join(data_dir, "raw", fn), skiprows=19, error_bad_lines=False
        )
        if df.shape[1] == 2:
            df.columns = ["date", "flow"]
        else:
            df.columns = ["date", "flow", "missing"]
        df["site"] = river
        flows.append(df)

    flows = pd.concat(flows)
    flows = flows.pivot(columns="site", index="date", values="flow")
    flows.index = pd.to_datetime(flows.index)

    wq_data = pd.read_csv(
        os.path.join(
            data_dir, "raw", "CEHThamesInitiative_WaterQualityData_2009-2013.csv"
        )
    )

    rain = pd.read_csv(
        os.path.join(data_dir, "raw", "39008_cdr.csv"),
        skiprows=19,
        error_bad_lines=False,
    )
    rain.columns = ["date", "value", "misc"]
    rain["site"] = "oxford_land"
    rain["variable"] = "precipitation"
    rain = rain.drop("misc", axis=1)
    rain.date = pd.to_datetime(rain.date)

    sites = {
        "River Thames at Swinford": "thames",
        "River Evenlode at Cassington Mill": "evenlode",
        "River Ray at Islip": "ray",
        "River Cherwell at Hampton Poyle": "cherwell",
    }
    wq = wq_data.loc[wq_data.Site.isin(sites.keys())]
    wq["Sampling date (dd/mm/yyyy)"] = pd.to_datetime(
        wq["Sampling date (dd/mm/yyyy)"], format="%d/%m/%Y"
    )
    wq = wq.rename(columns={"Sampling date (dd/mm/yyyy)": "date", "Site": "site"})

    wq.site = [sites[x] for x in wq.site]

    wq = wq.set_index("date").drop("Sampling time (hh:mm)", axis=1)
    wq[wq.columns.drop("site")] = wq[wq.columns.drop("site")].apply(
        lambda x: pd.to_numeric(x, errors="coerce")
    )
    wq = wq.dropna(axis=1, how="any")
    wq = wq.drop("Mean daily river discharge (m3 s-1)", axis=1)
    wq = wq.groupby("site").resample("d").interpolate().drop("site", axis=1)
    wq = wq.reset_index()
    wq.loc[:, wq.columns.str.contains("ug")] *= UG_L_TO_KG_M3
    wq.loc[:, wq.columns.str.contains("mg")] *= MG_L_TO_KG_M3

    columns = []
    for pol in wq.columns.unique():
        text = pol.lower()
        for sub in ["water", "dissolved", "total", " ", r"\(.*\)"]:
            text = re.sub(sub, "", text)
        columns.append(text)
    wq.columns = columns

    # Convert to nitrate as N
    wq["nitrate"] /= 4.43

    # Convert to Silica as SiO2
    wq["silicon"] *= 2.14

    wq = wq.melt(id_vars=["site", "date"])

    # wq.date = pd.to_datetime(wq.date.dt.date)

    flows = flows.loc[flows.index.isin(wq.date)]
    flows = flows.unstack().rename("value").reset_index()
    flows["variable"] = "flow"

    rain = rain.loc[rain.date.isin(wq.date)]
    evaporation = create_timeseries(2 / 1000, rain.date, "et0")
    evaporation["site"] = "oxford_land"

    temperature_ = (
        wq.loc[wq.variable == "temperature"].groupby("date").mean().reset_index()
    )
    temperature_["site"] = "oxford_land"
    temperature_["variable"] = "temperature"

    input_data = pd.concat([wq, flows, rain, evaporation, temperature_], axis=0)
    input_data.to_csv(
        os.path.join(data_dir, "processed", "timeseries_data.csv"), index=False
    )


def create_oxford_model(data_folder):
    """

    Args:
        data_folder:

    Returns:

    """
    input_fid = os.path.join(data_folder, "processed", "timeseries_data.csv")
    input_data = pd.read_csv(input_fid)
    input_data.loc[input_data.variable == "flow", "value"] *= constants.M3_S_TO_M3_DT
    input_data.loc[input_data.variable == "precipitation", "value"] *= constants.MM_TO_M
    input_data.date = pd.to_datetime(input_data.date)
    data_input_dict = input_data.set_index(["variable", "date"]).value.to_dict()
    data_input_dict = (
        input_data.groupby("site")
        .apply(lambda x: x.set_index(["variable", "date"]).value.to_dict())
        .to_dict()
    )
    dates = input_data.date.unique()
    dates = dates[dates.argsort()]
    dates = [pd.Timestamp(x) for x in dates]
    constants.POLLUTANTS = input_data.variable.unique().tolist()
    constants.POLLUTANTS.remove("flow")
    constants.POLLUTANTS.remove("precipitation")
    constants.POLLUTANTS.remove("et0")
    constants.NON_ADDITIVE_POLLUTANTS = ["temperature"]
    constants.ADDITIVE_POLLUTANTS = list(
        set(constants.POLLUTANTS).difference(constants.NON_ADDITIVE_POLLUTANTS)
    )
    constants.FLOAT_ACCURACY = 1e-8
    thames_above_abingdon = Waste(name="thames_above_abingdon")
    farmoor_abstraction = Node(name="farmoor_abstraction")
    evenlode_thames = Node(name="evenlode_thames")
    cherwell_ray = Node(name="cherwell_ray")
    cherwell_thames = Node(name="cherwell_thames")
    thames_mixer = Node(name="thames_mixer")
    evenlode = Catchment(name="evenlode", data_input_dict=data_input_dict["evenlode"])
    thames = Catchment(name="thames", data_input_dict=data_input_dict["thames"])
    ray = Catchment(name="ray", data_input_dict=data_input_dict["ray"])
    cherwell = Catchment(name="cherwell", data_input_dict=data_input_dict["cherwell"])
    oxford_fwtw = FWTW(
        service_reservoir_storage_capacity=1e5,
        service_reservoir_storage_area=2e4,
        service_reservoir_initial_storage=0.9e5,
        treatment_throughput_capacity=4.5e4,
        name="oxford_fwtw",
    )
    land_inputs = data_input_dict["oxford_land"]
    pollutant_deposition = {
        "boron": 100e-10,
        "calcium": 70e-7,
        "chloride": 60e-10,
        "fluoride": 0.2e-7,
        "magnesium": 6e-7,
        "nitrate": 2e-9,
        "nitrogen": 4e-7,
        "potassium": 7e-7,
        "silicon": 7e-9,
        "sodium": 30e-9,
        "sulphate": 70e-7,
    }
    surface = [
        {
            "type_": "PerviousSurface",
            "area": 2e7,
            "pollutant_load": pollutant_deposition,
            "surface": "rural",
            "field_capacity": 0.3,
            "depth": 0.5,
            "initial_storage": 2e7 * 0.4 * 0.5,
        },
        {
            "type_": "ImperviousSurface",
            "area": 1e7,
            "pollutant_load": pollutant_deposition,
            "surface": "urban",
            "initial_storage": 5e6,
        },
    ]
    oxford_land = Land(
        surfaces=surface, name="oxford_land", data_input_dict=land_inputs
    )
    oxford = ResidentialDemand(
        name="oxford",
        population=2e5,
        per_capita=0.15,
        pollutant_load={
            "boron": 500 * constants.UG_L_TO_KG_M3 * 0.15,
            "calcium": 150 * constants.MG_L_TO_KG_M3 * 0.15,
            "chloride": 180 * constants.MG_L_TO_KG_M3 * 0.15,
            "fluoride": 0.4 * constants.MG_L_TO_KG_M3 * 0.15,
            "magnesium": 30 * constants.MG_L_TO_KG_M3 * 0.15,
            "nitrate": 60 * constants.MG_L_TO_KG_M3 * 0.15,
            "nitrogen": 50 * constants.MG_L_TO_KG_M3 * 0.15,
            "potassium": 30 * constants.MG_L_TO_KG_M3 * 0.15,
            "silicon": 20 * constants.MG_L_TO_KG_M3 * 0.15,
            "sodium": 200 * constants.MG_L_TO_KG_M3 * 0.15,
            "sulphate": 250 * constants.MG_L_TO_KG_M3 * 0.15,
            "temperature": 14,
        },
        data_input_dict=land_inputs,
    )
    farmoor = Reservoir(
        name="farmoor", capacity=1e7, initial_storage=1e7, area=1.5e6, datum=62
    )
    distribution = Node(name="oxford_distribution")
    oxford_wwtw = WWTW(
        stormwater_storage_capacity=2e4,
        stormwater_storage_area=2e4,
        treatment_throughput_capacity=5e4,
        name="oxford_wwtw",
    )
    combined_sewer = Sewer(
        capacity=4e6, pipe_timearea={0: 0.8, 1: 0.15, 2: 0.05}, name="combined_sewer"
    )
    gw = Groundwater(capacity=3.2e9, area=3.2e8, name="gw", residence_time=20)
    nodelist = [
        thames_above_abingdon,
        evenlode,
        thames,
        ray,
        cherwell,
        oxford,
        distribution,
        farmoor,
        oxford_fwtw,
        oxford_wwtw,
        combined_sewer,
        oxford_land,
        gw,
        farmoor_abstraction,
        evenlode_thames,
        cherwell_ray,
        cherwell_thames,
        thames_mixer,
    ]
    fwtw_to_distribution = Arc(
        in_port=oxford_fwtw, out_port=distribution, name="fwtw_to_distribution"
    )
    abstraction_to_farmoor = Arc(
        in_port=farmoor_abstraction,
        out_port=farmoor,
        name="abstraction_to_farmoor",
        capacity=5e4,
    )
    sewer_to_wwtw = Arc(
        in_port=combined_sewer,
        out_port=oxford_wwtw,
        preference=1e10,
        name="sewer_to_wwtw",
    )
    sewer_overflow = Arc(
        in_port=combined_sewer,
        out_port=thames_mixer,
        preference=1e-10,
        name="sewer_overflow",
    )
    evenlode_to_thames = Arc(
        in_port=evenlode, out_port=evenlode_thames, name="evenlode_to_thames"
    )

    thames_to_thames = Arc(
        in_port=thames, out_port=evenlode_thames, name="thames_to_thames"
    )

    ray_to_cherwell = Arc(in_port=ray, out_port=cherwell_ray, name="ray_to_cherwell")

    cherwell_to_cherwell = Arc(
        in_port=cherwell, out_port=cherwell_ray, name="cherwell_to_cherwell"
    )

    thames_to_farmoor = Arc(
        in_port=evenlode_thames, out_port=farmoor_abstraction, name="thames_to_farmoor"
    )

    farmoor_to_mixer = Arc(
        in_port=farmoor_abstraction, out_port=thames_mixer, name="farmoor_to_mixer"
    )

    cherwell_to_mixer = Arc(
        in_port=cherwell_ray, out_port=thames_mixer, name="cherwell_to_mixer"
    )

    wwtw_to_mixer = Arc(
        in_port=oxford_wwtw, out_port=thames_mixer, name="wwtw_to_mixer"
    )

    mixer_to_waste = Arc(
        in_port=thames_mixer, out_port=thames_above_abingdon, name="mixer_to_waste"
    )

    distribution_to_demand = Arc(
        in_port=distribution, out_port=oxford, name="distribution_to_demand"
    )

    reservoir_to_fwtw = Arc(
        in_port=farmoor, out_port=oxford_fwtw, name="reservoir_to_fwtw"
    )

    fwtw_to_sewer = Arc(
        in_port=oxford_fwtw, out_port=combined_sewer, name="fwtw_to_sewer"
    )

    demand_to_sewer = Arc(
        in_port=oxford, out_port=combined_sewer, name="demand_to_sewer"
    )

    land_to_sewer = Arc(
        in_port=oxford_land, out_port=combined_sewer, name="land_to_sewer"
    )

    land_to_gw = Arc(in_port=oxford_land, out_port=gw, name="land_to_gw")

    garden_to_gw = Arc(in_port=oxford, out_port=gw, name="garden_to_gw")

    gw_to_mixer = Arc(in_port=gw, out_port=thames_mixer, name="gw_to_mixer")
    arclist = [
        evenlode_to_thames,
        thames_to_thames,
        ray_to_cherwell,
        cherwell_to_cherwell,
        thames_to_farmoor,
        farmoor_to_mixer,
        cherwell_to_mixer,
        wwtw_to_mixer,
        sewer_overflow,
        mixer_to_waste,
        abstraction_to_farmoor,
        distribution_to_demand,
        demand_to_sewer,
        land_to_sewer,
        sewer_to_wwtw,
        fwtw_to_sewer,
        fwtw_to_distribution,
        reservoir_to_fwtw,
        land_to_gw,
        garden_to_gw,
        gw_to_mixer,
    ]
    my_model = Model()
    my_model.add_instantiated_nodes(nodelist)
    my_model.add_instantiated_arcs(arclist)
    my_model.dates = dates
    return my_model


def pull_check_mrf(node, vqip=None):
    """

    Args:
        node:
        vqip:

    Returns:

    """
    reply = node.pull_check_basic()
    reply["volume"] = max(
        reply["volume"] - (node.mrf - node.mrf_satisfied_this_timestep), 0
    )
    if vqip is not None:
        reply["volume"] = min(reply["volume"], vqip["volume"])
    return reply


def pull_set_mrf(node, vqip):
    """

    Args:
        node:
        vqip:

    Returns:

    """
    request = node.copy_vqip(vqip)
    request["volume"] += node.mrf - node.mrf_satisfied_this_timestep
    reply = node.pull_distributed(request)
    reply_to_mrf = min((node.mrf - node.mrf_satisfied_this_timestep), reply["volume"])
    node.mrf_satisfied_this_timestep += reply_to_mrf
    reply_to_mrf = node.v_change_vqip(reply, reply_to_mrf)
    reply = node.extract_vqip(reply, reply_to_mrf)
    mrf_route_reply = node.push_distributed(reply_to_mrf, of_type=["Node"])
    if mrf_route_reply["volume"] > constants.FLOAT_ACCURACY:
        print("warning MRF not able to push")
    return reply


def push_set_mrf(node, vqip):
    """

    Args:
        node:
        vqip:

    Returns:

    """
    reply = node.push_distributed(vqip)
    total_pushed_downstream = vqip["volume"] - reply["volume"]
    node.mrf_satisfied_this_timestep = min(
        node.mrf, node.mrf_satisfied_this_timestep + total_pushed_downstream
    )
    return reply


def end_timestep(node):
    """

    Args:
        node:
    """
    node.mrf_satisfied_this_timestep = 0


def convert_to_mrf(node, mrf=5):
    """

    Args:
        node:
        mrf:
    """
    node.mrf = mrf
    node.mrf_satisfied_this_timestep = 0
    node.pull_set_handler["default"] = lambda x: pull_set_mrf(node, x)
    node.pull_check_handler["default"] = lambda x: pull_check_mrf(node, x)
    node.end_timestep = lambda: end_timestep(node)


def create_oxford_model_mrf(data_folder):
    """

    Args:
        data_folder:

    Returns:

    """
    my_model = create_oxford_model(data_folder)
    convert_to_mrf(my_model.nodes["farmoor_abstraction"], mrf=3 * 86400)
    return my_model


def create_timeseries(amount, dates, variable):
    """Create a timeseries with a constant value, formatted as a dataframe.

    Args:
        amount (float): Constant value to be applied over the timeseries
        dates (list): list or iterable of dates to be generated
        variable (str): String to store in the 'variable' column of the dataframe

    Returns:
        (DataFrame): the formatted dataframe
    """
    df = pd.DataFrame(columns=["date", "variable", "value"])
    df["date"] = dates
    df["variable"] = variable
    df["value"] = amount
    return df



"""Tests for the tanks module."""

import unittest
from unittest import TestCase

from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.tanks import (
    DecayQueueTank,
    DecayTank,
    QueueTank,
    ResidenceTank,
    Tank,
)


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_tank_ds(self):
        tank = Tank(
            capacity=10,
            initial_storage={"volume": 5, "phosphate": 0.4, "temperature": 10},
        )
        tank.end_timestep()

        d1 = {"volume": 2, "phosphate": 0.01, "temperature": 15}

        _ = tank.push_storage(d1)

        diff = tank.ds()

        d2 = {"volume": 2, "phosphate": 0.01, "temperature": 0}

        self.assertDictAlmostEqual(d2, diff, 16)

    def test_ponded(self):
        tank = Tank(
            capacity=10,
            initial_storage={"volume": 15, "phosphate": 0.4, "temperature": 10},
        )
        d1 = {"volume": 5, "phosphate": 0.4 / 3, "temperature": 10}
        reply = tank.pull_ponded()
        self.assertDictAlmostEqual(d1, reply)

    def test_tank_get_avail(self):
        d1 = {"volume": 5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        reply = tank.get_avail()
        self.assertDictAlmostEqual(d1, reply)

        reply = tank.get_avail({"volume": 2.5})
        d2 = {"volume": 2.5, "phosphate": 0.2, "temperature": 10}
        self.assertDictAlmostEqual(d2, reply)

        reply = tank.get_avail({"volume": 10})
        self.assertDictAlmostEqual(d1, reply)

    def test_tank_get_excess(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        d2 = {"volume": 2.5, "phosphate": 0.4 / 3, "temperature": 10}
        reply = tank.get_excess()
        self.assertDictAlmostEqual(d2, reply)

        d2 = {"volume": 1, "phosphate": 0.4 * 1 / 7.5, "temperature": 10}
        reply = tank.get_excess({"volume": 1})
        self.assertDictAlmostEqual(d2, reply)

    def test_tank_push_storage(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        d2 = {"volume": 5, "phosphate": 0.4, "temperature": 15}

        d3 = {"volume": 2.5, "phosphate": 0.2, "temperature": 15}
        reply = tank.push_storage(d2)
        self.assertDictAlmostEqual(d3, reply)

        d4 = {"volume": 0, "phosphate": 0, "temperature": 0}
        reply = tank.push_storage(d2, force=True)
        self.assertDictAlmostEqual(d4, reply)

    def test_tank_pull_storage(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        d2 = {"volume": 5, "phosphate": 0.4 * 5 / 7.5, "temperature": 10}

        reply = tank.pull_storage({"volume": 5})
        self.assertDictAlmostEqual(d2, reply)

        d3 = {"volume": 2.5, "phosphate": 0.4 * 2.5 / 7.5, "temperature": 10}

        reply = tank.pull_storage({"volume": 5})

        self.assertDictAlmostEqual(d3, reply, 15)

    def test_tank_pull_pollutants(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        d2 = {"volume": 5, "phosphate": 0.1, "temperature": 10}

        reply = tank.pull_pollutants(d2)
        self.assertDictAlmostEqual(d2, reply)

        reply = tank.pull_pollutants(d2)
        d3 = {"volume": 2.5, "phosphate": 0.1, "temperature": 10}
        self.assertDictAlmostEqual(d3, reply, 15)

    def test_tank_head(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1, datum=5, area=2.5)

        reply = tank.get_head()
        self.assertEqual(8, reply)

        reply = tank.get_head(datum=-1)
        self.assertEqual(2, reply)

        reply = tank.get_head(non_head_storage=2)
        self.assertEqual(7.2, reply)

        reply = tank.get_head(non_head_storage=10)
        self.assertEqual(5, reply)

    def test_evap(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = Tank(capacity=10, initial_storage=d1)

        d2 = {"volume": 0, "phosphate": 0.4, "temperature": 10}

        reply = tank.evaporate(10)
        self.assertEqual(7.5, reply)
        self.assertDictAlmostEqual(d2, tank.storage)

    def test_residence_tank(self):
        d1 = {"volume": 7.5, "phosphate": 0.4, "temperature": 10}
        tank = ResidenceTank(residence_time=3, initial_storage=d1)

        d2 = {"volume": 2.5, "phosphate": 0.4 / 3, "temperature": 10}
        reply = tank.pull_outflow()
        self.assertDictAlmostEqual(d2, reply)

    def test_decay_tank(self):
        node = Node(name="", data_input_dict={("temperature", 1): 15})
        node.t = 1
        d1 = {"volume": 8, "phosphate": 0.4, "temperature": 10}

        tank = DecayTank(
            decays={"phosphate": {"constant": 0.001, "exponent": 1.005}},
            initial_storage=d1,
            parent=node,
        )
        _ = tank.pull_storage({"volume": 2})

        d3 = {"volume": -2, "phosphate": -0.1, "temperature": 0}

        diff = tank.decay_ds()
        self.assertDictAlmostEqual(d3, diff, 16)

        tank.end_timestep_decay()

        d2 = {
            "volume": 6,
            "phosphate": 0.3 - 0.3 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 10,
        }

        self.assertDictAlmostEqual(d2, tank.storage, 16)

        self.assertAlmostEqual(
            0.3 * 0.001 * 1.005 ** (15 - 20), tank.total_decayed["phosphate"]
        )

    def test_queue_push(self):
        d1 = {"volume": 5, "phosphate": 0.4, "temperature": 10}
        tank = QueueTank(number_of_timesteps=1, capacity=10, initial_storage=d1)

        d2 = {"volume": 1, "phosphate": 0.1, "temperature": 15}

        tank.push_storage(d2)

        d3 = {"volume": 6, "phosphate": 0.5, "temperature": (5 * 10 + 15) / 6}

        self.assertDictAlmostEqual(d3, tank.storage)
        self.assertDictAlmostEqual(d1, tank.active_storage)
        self.assertDictAlmostEqual(d2, tank.internal_arc.queue[1])

        tank.push_storage(d2, force=True)
        self.assertDictAlmostEqual(d3, tank.active_storage)

        tank.end_timestep()

        d4 = {"volume": 7, "phosphate": 0.6, "temperature": ((5 * 10) + (15 * 2)) / 7}
        self.assertDictAlmostEqual(d4, tank.active_storage)

    def test_queue_pull(self):
        d1 = {"volume": 5, "phosphate": 0.4, "temperature": 10}
        tank = QueueTank(number_of_timesteps=1, capacity=10, initial_storage=d1)
        d2 = {"volume": 1, "phosphate": 0.1, "temperature": 15}
        reply = tank.push_storage(d2)

        reply = tank.pull_storage({"volume": 6})
        self.assertDictAlmostEqual(d1, reply)
        tank.end_timestep()
        self.assertDictAlmostEqual(d2, tank.active_storage)

    def test_queue_pull_exact(self):
        d1 = {"volume": 5, "phosphate": 0.4, "temperature": 10}
        tank = QueueTank(number_of_timesteps=1, capacity=10, initial_storage=d1)
        d2 = {"volume": 1, "phosphate": 0.1, "temperature": 15}
        reply = tank.push_storage(d2)

        reply = tank.pull_storage_exact(
            {"volume": 6, "phosphate": 0.1, "temperature": 10}
        )

        d3 = {"volume": 5, "phosphate": 0.1, "temperature": 10}
        self.assertDictAlmostEqual(d3, reply)

        reply = tank.pull_storage_exact(
            {"volume": 0, "phosphate": 0.6, "temperature": 10}
        )
        d4 = {"volume": 0, "phosphate": 0.3, "temperature": 10}
        self.assertDictAlmostEqual(d4, reply, 16)

    def test_decay_queue(self):
        node = Node(name="", data_input_dict={("temperature", 1): 15})
        node.t = 1
        d1 = {"volume": 5, "phosphate": 0.4, "temperature": 10}
        tank = DecayQueueTank(
            number_of_timesteps=1,
            capacity=10,
            initial_storage=d1,
            decays={"phosphate": {"constant": 0.001, "exponent": 1.005}},
            parent=node,
        )

        d2 = {"volume": 1, "phosphate": 0.1, "temperature": 15}

        _ = tank.push_storage(d2)

        tank.end_timestep()

        d4 = {
            "volume": 6,
            "phosphate": 0.4 + 0.1 * (1 - 0.001 * 1.005 ** (15 - 20)),
            "temperature": ((5 * 10) + (15 * 1)) / 6,
        }
        self.assertDictAlmostEqual(d4, tank.storage, 15)

    def test_overrides(self):
        # node - no need to test
        # tank
        tank = Tank(capacity=10, area=8, datum=4)
        tank.apply_overrides({"capacity": 3, "area": 2, "datum": 3.5})
        self.assertEqual(tank.capacity, 3)
        self.assertEqual(tank.area, 2)
        self.assertEqual(tank.datum, 3.5)
        # residence tank
        tank = ResidenceTank(capacity=10, area=8, datum=4, residence_time=8)
        tank.apply_overrides(
            {"capacity": 3, "area": 2, "datum": 3.5, "residence_time": 6}
        )
        self.assertEqual(tank.capacity, 3)
        self.assertEqual(tank.area, 2)
        self.assertEqual(tank.datum, 3.5)
        self.assertEqual(tank.residence_time, 6)
        # decay tank
        tank = DecayTank(
            capacity=10,
            area=8,
            datum=4,
            decays={"nitrate": {"constant": 0.001, "exponent": 1.005}},
        )
        tank.apply_overrides(
            {
                "capacity": 3,
                "area": 2,
                "datum": 3.5,
                "decays": {"phosphate": {"constant": 1.001, "exponent": 10.005}},
            }
        )
        self.assertEqual(tank.capacity, 3)
        self.assertEqual(tank.area, 2)
        self.assertEqual(tank.datum, 3.5)
        self.assertDictEqual(
            tank.decays,
            {
                "nitrate": {"constant": 0.001, "exponent": 1.005},
                "phosphate": {"constant": 1.001, "exponent": 10.005},
            },
        )
        # queue tank
        tank = QueueTank(capacity=10, area=8, datum=4, number_of_timesteps=8)
        tank.apply_overrides(
            {"capacity": 3, "area": 2, "datum": 3.5, "number_of_timesteps": 6}
        )
        self.assertEqual(tank.capacity, 3)
        self.assertEqual(tank.area, 2)
        self.assertEqual(tank.datum, 3.5)
        self.assertEqual(tank.number_of_timesteps, 6)
        self.assertEqual(tank.internal_arc.number_of_timesteps, 6)
        # decay queue tank
        tank = DecayQueueTank(
            capacity=10,
            area=8,
            datum=4,
            number_of_timesteps=8,
            decays={"phosphate": {"constant": 0.001, "exponent": 1.005}},
        )
        tank.apply_overrides(
            {
                "capacity": 3,
                "area": 2,
                "datum": 3.5,
                "number_of_timesteps": 6,
                "decays": {"phosphate": {"constant": 1.001, "exponent": 10.005}},
            }
        )
        self.assertEqual(tank.capacity, 3)
        self.assertEqual(tank.area, 2)
        self.assertEqual(tank.datum, 3.5)
        self.assertEqual(tank.number_of_timesteps, 6)
        self.assertEqual(tank.internal_arc.number_of_timesteps, 6)
        self.assertDictEqual(
            tank.internal_arc.decays,
            {"phosphate": {"constant": 1.001, "exponent": 10.005}},
        )


if __name__ == "__main__":
    unittest.main()
